import duckdb
import sqlite3
import time
import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt

def read_id_file(fname):
    with open(fname, 'r') as f:
        return list(f.readlines())

TRIP_IDS = read_id_file('data/trip_ids.txt')
TRIP_ROUTE_IDS = read_id_file('data/trip_route_ids.txt')
STOP_TIMES_IDS = read_id_file('data/stop_times_ids.txt')

class DuckDB:
    """Interface for using DuckDB"""
    def __init__(self):
        duckdb.sql("CREATE TABLE trips AS SELECT * FROM read_csv_auto('data/trips.txt')")
        duckdb.sql("CREATE UNIQUE INDEX trips_id_idx ON trips(trip_id)")

        duckdb.sql("CREATE TABLE stop_times AS SELECT * FROM read_csv_auto('data/stop_times.txt')")
        duckdb.sql("CREATE UNIQUE INDEX stop_times_id_idx ON stop_times(trip_id,stop_sequence)")

    def query(self, stmt):
        return duckdb.sql(stmt)

class SQLiteDB:
    """Interface for using SQLite"""
    def __init__(self):
        self._db = sqlite3.connect('data/gtfs.sqlite3')
        self._cursor = self._db.cursor()

    def query(self, stmt):
        return self._cursor.execute(stmt).fetchall()

    def raw_query(self, stmt):
        return self._cursor.execute(stmt)
    

class Benchmarker:
    """Run a list of SQL queries against both DuckDB and SQLite.

    Also provides plotting functionality."""
    def __init__(self):
        self._data = []
        self._ddb = DuckDB()
        self._sdb = SQLiteDB()

    def _run_ddb(self, name, querylist, count):
        # warm up the cache
        for stmt in querylist:
            self._ddb.query(stmt)

        t1 = time.time()
        rescount = 0
        for stmt in querylist:
            rescount += len(self._ddb.query(stmt))
        t2 = time.time() 

        self._data.append({'name': name,
                           'duration': t2-t1,
                           'duration_per_query': ((t2 - t1) / count) * 1000.0,
                           'db': 'duck',
                           'rescount': rescount,
                           'count': count})

    def _run_sdb(self, name, querylist, count):
        # warm up the caches
        for stmt in querylist:
            self._sdb.query(stmt)
        
        t1 = time.time()
        rescount = 0
        for stmt in querylist:
            rescount += len(self._sdb.query(stmt))
        t2 = time.time()

        self._data.append({'name': name,
                           'duration': t2-t1,
                           'duration_per_query': ((t2 - t1) / count) * 1000.0,
                           'db': 'sqlite',
                           'rescount': rescount,
                           'count': count})
        
    def run(self, name, querylist, count):
        self._run_ddb(name, querylist, count)
        self._run_sdb(name, querylist, count)
        
    def print(self):
        for d in self._data:
            print(f"{d['name']:>25} | {d['db']:>6} | {d['duration_per_query']:>6}")

    def plot(self, fname, benchmarks):
        full_df = pd.DataFrame(self._data)
        df = full_df.loc[full_df['name'].isin(benchmarks)]
        g = sns.catplot(
            data=df, kind="bar",
            x="name", y="duration_per_query", hue="db",
            palette="dark", alpha=.6, height=6
        )
        g.despine(left=True)
        g.set_axis_labels("", "Time per Query (ms)")
        g.legend.set_title("")
        g.savefig(fname)

b = Benchmarker()

def retr_trips(count):
    stmts = []
    for trip_id in TRIP_IDS[:count]:
        stmts.append(f'SELECT * FROM trips WHERE trip_id = {trip_id};')
    b.run(f'indexed', stmts, count)

def retr_trips_by_route_id(count):
    stmts = []
    for route_id in TRIP_ROUTE_IDS[:count]:
        stmts.append(f"SELECT * FROM trips WHERE route_id = '{route_id}';")
    b.run(f'nonindexed', stmts, count)

def retr_trips_by_dir_id(count):
    stmts = []
    for i in range(0, count):
        dir_id = i % 2
        stmts.append(f'SELECT * FROM trips WHERE direction_id = {dir_id};')
    b.run(f'many', stmts, count)

def retr_stop_times_by_id(count):
    stmts = []
    for stop_times_id in STOP_TIMES_IDS[:count]:
        [trip_id, stop_sequence] = stop_times_id.split('|')
        stmts.append(f'SELECT * FROM stop_times WHERE trip_id = {trip_id} and stop_sequence = {stop_sequence};')
    b.run(f'composite-indexed', stmts, count)

def retr_join(count):
    stmts = []
    for stop_times_id in STOP_TIMES_IDS[:count]:
        [trip_id, stop_sequence] = stop_times_id.split('|')
        stmts.append(f'SELECT * FROM stop_times JOIN trips ON trips.trip_id = stop_times.trip_id WHERE stop_times.trip_id = {trip_id} and stop_sequence = {stop_sequence};')
    b.run(f'join', stmts, count)

def retr_trips_multiple(count, multicount):
    stmts = []
    def chunker(seq, size):
        return (seq[pos:pos + size] for pos in range(0, len(seq), size))
    for chunk in list(chunker(TRIP_IDS, multicount))[:count]:
        stmt = "SELECT * from TRIPS where trip_id in ("
        first = True
        for trip_id in chunk:
            if not first:
                stmt += f",{trip_id}"
            else:
                stmt += f"{trip_id}"
            first = False
        stmt += ");"
        stmts.append(stmt)
    b.run(f'indexed-multiple', stmts, count)
            

# List of (benchmark_function, count)
RUNS = [
    (retr_stop_times_by_id,2000),
    (retr_join,2000),
    (retr_trips,2000),
    (retr_trips_by_route_id,200),
    (retr_trips_by_dir_id,200)
]

for (fun,count) in RUNS:
    fun(count)

# this one is special since it needs to know how many to retrieve in one go
retr_trips_multiple(2000,50)
        
b.print()
b.plot('/tmp/dbench_plot_indexed.pdf', ['indexed', 'composite-indexed', 'join', 'indexed-multiple'])
b.plot('/tmp/dbench_plot_nonindexed.pdf', ['nonindexed', 'many'])
